# cython: language_level=3
"""
Contains the functions needed for computing the precessing post-adiabatic dynamics
"""

cimport cython
from typing import Dict
import numpy as np
import  quaternion
from libc.math cimport log, sqrt, exp, abs,fabs, tgamma,sin,cos, tanh, sinh, asinh

from .initial_conditions_aligned_opt import computeIC_opt as computeIC
from ..fits.fits_Hamiltonian import dSO as dSO_poly_fit

from pygsl_lite import  roots, errno
from scipy import optimize
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline
from scipy import integrate

from lalinference.imrtgr import nrutils

from pyseobnr.eob.utils.containers cimport EOBParams
from pyseobnr.eob.hamiltonian.Hamiltonian_v5PHM_C cimport Hamiltonian_v5PHM_C
from pyseobnr.eob.waveform.waveform cimport RadiationReactionForce

from .initial_conditions_aligned_precessing import computeIC_augm
from .pn_evolution_opt import compute_quasiprecessing_PNdynamics_opt
from .pn_evolution_opt import build_splines_PN
from .integrate_ode_prec import compute_dynamics_prec_opt
from pyseobnr.eob.utils.math_ops_opt import my_dot, my_norm

from .postadiabatic_C import Kerr_ISCO, Newtonian_j0, univariate_spline_integral

@cython.profile(True)
@cython.linetrace(True)
cpdef precessing_final_spin(
    double chi1_LN,
    double chi2_LN,
    chi1_v,
    chi2_v,
    LNhat,
    double m1,
    double m2,
):
    """
    Magnitude of the final spin computed from NR fits of precessing NR simulations. The sign of the final spin is computed
    from the non-precessing NR fits.

    Args:
        chi1_LN (double): Projection of the primary dimensionless spin onto LNhat
        chi2_LN (double): Projection of the secondary dimensionless spin onto LNhat
        chi1_v (double[:]): Primary dimensionless spin vector
        chi2_v (double[:]): Secondary dimensionless spin vector
        LNhat (double[:]): Newtonian orbital angular momentum vector
        m1 (double): Mass of the primary
        m2 (double): Mass of the secondary

    Returns:
        (double) Final spin
    """
    # Spin magnitudes
    cdef double a1 = my_norm(chi1_v)
    cdef double a2 = my_norm(chi2_v)

    cdef double tilt1, tilt2, angle, phi12

    # Tilt angles w.r.t L_N
    if a1 != 0:
        angle = chi1_LN / a1
        if angle < -1:
            tilt1 = np.pi
        elif angle > 1:
            tilt1 = 0.0
        else:
            tilt1 = np.arccos(np.clip(angle,-1,1))
    else:
        tilt1 = 0.0

    if a2 != 0:
        angle = chi2_LN / a2
        if angle < -1:
            tilt2 = np.pi
        elif angle > 1:
            tilt2 = 0.0
        else:
            tilt2 = np.arccos(np.clip(angle,-1,1))
    else:
        tilt2 = 0.0

    # Angle between the inplane spin components
    chi1_perp = chi1_v - chi1_LN * LNhat
    chi2_perp = chi2_v - chi2_LN * LNhat

    if a1 == 0 or a2 == 0:
        phi12 = np.pi / 2.0

    else:
        angle = np.dot(chi1_perp / a1, chi2_perp / a2)
        phi12 = np.arccos(np.clip(angle,-1,1))

    # Call non-precessing HBR fit to get the *sign* of the final spin
    cdef double final_spin_nonprecessing = nrutils.bbh_final_spin_non_precessing_HBR2016(
        m1, m2, chi1_LN, chi2_LN, version="M3J4"
    )
    final_spin_noprec = final_spin_nonprecessing
    cdef double sign_final_spin = +1

    cdef double norm_final_spin_nonprecessing = np.linalg.norm( final_spin_nonprecessing)
    if norm_final_spin_nonprecessing !=0:
      sign_final_spin = final_spin_nonprecessing / norm_final_spin_nonprecessing

    # Compute the magnitude of the final spin using the precessing fit
    final_spin = nrutils.bbh_final_spin_precessing_HBR2016(
        m1, m2, a1, a2, tilt1, tilt2, phi12, version="M3J4"
    )

    # Flip sign if needed
    final_spin *= sign_final_spin

    return final_spin


@cython.profile(True)
@cython.linetrace(True)
cpdef double j0_eqn(
    double j0_sol,
    double r,
    Hamiltonian_v5PHM_C H,
    double[:] chi1_v,
    double[:] chi2_v,
    double m_1,
    double m_2,
    double chi1_LN,
    double chi2_LN,
    double chi1_L,
    double chi2_L,
    EOBParams params,
    double[:] q,
    double[:] p,
):
    """
    Equation dH_dr =0 evaluated at prstar=0 used to obtain the adiabatic solution
    for the orbital angular momentum (pphi)


    Args:
        j0_sol (double): Guess of the solution
        r (double): Orbital separation
        H (Hamiltonian_v5PHM_C): Hamiltonian
        chi1_v (double[:]): Primary dimensionless spin vector
        chi2_v (double[:]): Secondary dimensionless spin vector
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        chi1_LN (double): Projection of the primary dimensionless spin onto LNhat
        chi2_LN (double): Projection of the secondary dimensionless spin onto LNhat
        chi1_L (double): Projection of the primary dimensionless spin onto Lhat
        chi2_L (double): Projection of the secondary dimensionless spin onto Lhat
        params (EOBParams): Container with useful EOB parameters
        q (double[:]): Position q=[r,phi]
        p (double[:]): Canonical momentum p=[pr,pphi]

    Returns:
        (double) Value of the orbital angular momemtum (pphi)
    """

    q[0] = r
    p[1]= j0_sol
    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double ap = chi1_LN * X1 + chi2_LN * X2
    cdef double am = chi1_LN * X1 - chi2_LN * X2

    cdef double dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)

    H.calibration_coeffs['dSO'] = dSO_new
    cdef (double,double,double,double) dH_dq = H.grad(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )
    cdef double dH_dr = dH_dq[0]
    return dH_dr

cpdef compute_adiabatic_solution(
    double[:] r,
    double[:] omega,
    Hamiltonian_v5PHM_C H,
    dict splines,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    EOBParams params,
    double tol=1e-12,
):
    """
    Compute adiabatic solution for pphi by solving dH_dr =0 with prstar=0

    Args:
        r (double[:]): Radial grid on which to compute the adiabatic solution
        omega (double[:]): Grid of orbital frequencies used to compute the adiabatic solution
        H (Hamiltonian_v5PHM_C): Hamiltonian
        splines (dict): Dictionary containing the interpolated spin-precessing evolution
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.
        params (EOBParams): Container with useful EOB parameters
        tol (double): Tolerance for the find root

    Returns:
        (np.array) Values of adiabatic pphi along the radial grid
    """

    cdef int i
    cdef double[:] j0 = Newtonian_j0(r)

    cdef double omega_start = splines["everything"].x[0]

    tmp = splines["everything"](omega)

    cdef double[:] chi1_LN = tmp[:,0]
    cdef double[:] chi2_LN = tmp[:,1]

    cdef double[:] chi1_L = tmp[:,2]
    cdef double[:] chi2_L = tmp[:,3]

    chi1_v = tmp[:,4:7]
    chi2_v = tmp[:,7:10]


    for i in range(r.shape[0]):

        if omega[i] < 0.9*omega_start:
          print(f"problem PA dynamics is extrapolating: omega ={omega[i]}< 0.9*omega_start {omega_start}")
        j0_solution = optimize.root(
            j0_eqn,
            j0[i],
            args=(r[i], H, chi1_v[i], chi2_v[i],
            chi1_LN[i], chi2_LN[i],
            chi1_L[i], chi2_L[i],
            m_1, m_2,params,q,p),
            tol=tol,
        )
        j0[i] = j0_solution.x

    return j0

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pr_eqn(
    pr_sol,
    double r,
    double pphi,
    double dpphi_dr,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    chi1_v,
    chi2_v,
    double chi1_LN,
    double chi2_LN,
    double chi1_L,
    double chi2_L,
    double m_1,
    double m_2,
    double omega_start,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Equation dpphi_dr * dH_dpr * csi - flux[1] = 0, where flux[1] is the azimuthal
    component of the radiation reaction force and csi is the tortoise coordinate
    transformation factor, to solve for prstar in the numerical postadiabatic routine.

    Args:
        pr_sol: Solution for prstar from the previous post-adiabatic order
        r (double): Orbital separation
        omega (double): Orbital frequency
        pphi (double): Orbital angular momentum
        dpphi_dr (double): Radial derivative of the orbital angular momentum
        H (Hamiltonian_v5PHM_C): Hamiltonian
        RR (RadiationReactionForce): Radiation Reaction Force
        chi1_v : Primary dimensionless spin vector
        chi2_v : Secondary dimensionless spin vector
        chi1_LN (double): Projection of the primary dimensionless spin onto LNhat
        chi2_LN (double): Projection of the secondary dimensionless spin onto LNhat
        chi1_L (double): Projection of the primary dimensionless spin onto Lhat
        chi2_L (double): Projection of the secondary dimensionless spin onto Lhat
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        omega_start (double): Initial orbital frequency
        params (EOBParams): Container with useful EOB parameters
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.

    Returns:
        (double) Value of prstar
    """
    q[0] = r
    p[0] = pr_sol[0]
    p[1] = pphi
    cdef double dH_dpr,H_val,csi,omega,omega_circ,result
    cdef double dynamics[6]

    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double ap = chi1_LN * X1 + chi2_LN * X2
    cdef double am = chi1_LN * X1 - chi2_LN * X2

    cdef double dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)

    H.calibration_coeffs['dSO'] = dSO_new

    dynamics[:] = H.dynamics(q, p, chi1_v, chi2_v, m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )

    dH_dpr = dynamics[2]
    omega = dynamics[3]
    H_val = dynamics[4]
    csi = dynamics[5]

    params.dynamics.p_circ[1] = pphi
    omega_circ = H.omega(q,params.dynamics.p_circ,chi1_v,chi2_v,m_1,m_2,chi1_LN, chi2_LN,
        chi1_L, chi2_L
    )

    if omega < 0.9*omega_start:
      print(f"problem PA dynamics is extrapolating: omega ={omega}< 0.9*omega_start {omega_start}")

    params.p_params.update_spins(chi1_LN, chi2_LN)
    cdef (double,double) flux = RR.RR(q, p, omega, omega_circ, H_val, params)

    params.p_params.omega_circ = omega_circ
    params.p_params.omega = omega
    params.p_params.H_val = H_val

    result = dpphi_dr * dH_dpr * csi - flux[1]

    return result


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pr_eqn_analytic(
    pr_sol,
    double r,
    double pphi,
    double dpphi_dr,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    chi1_v,
    chi2_v,
    double chi1_LN,
    double chi2_LN,
    double chi1_L,
    double chi2_L,
    double m_1,
    double m_2,
    double omega_start,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Equation dpphi_dr * dH_dpr * csi - flux[1] = 0, where flux[1] is the azimuthal
    component of the radiation reaction force and csi is the tortoise coordinate
    transformation factor, to solve for prstar in the analytic postadiabatic routine.
    This function provides the same result as the function "pr_eqn", but instead of
    using a root finding routine it computes the roots analytically using the
    potentials of the Hamiltonian.

    Args:
        pr_sol: Solution for prstar from the previous post-adiabatic order
        r (double): Orbital separation
        pphi (double): Orbital angular momentum
        dpphi_dr (double): Radial derivative of the orbital angular momentum
        H (Hamiltonian_v5PHM_C): Hamiltonian
        RR (RadiationReactionForce): Radiation Reaction Force
        chi1_v : Primary dimensionless spin vector
        chi2_v : Secondary dimensionless spin vector
        chi1_LN (double): Projection of the primary dimensionless spin onto LNhat
        chi2_LN (double): Projection of the secondary dimensionless spin onto LNhat
        chi1_L (double): Projection of the primary dimensionless spin onto Lhat
        chi2_L (double): Projection of the secondary dimensionless spin onto Lhat
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        omega_start (double): Initial orbital frequency
        params (EOBParams): Container with useful EOB parameters
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.

    Returns:
        (double) Value of prstar
    """

    q[0] = r
    p[0] = pr_sol
    p[1] = pphi

    cdef double nu = params.p_params.nu

    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double ap = chi1_LN * X1 + chi2_LN * X2
    cdef double am = chi1_LN * X1 - chi2_LN * X2

    cdef double dSO_new = dSO_poly_fit(nu, ap, am)
    H.calibration_coeffs['dSO'] = dSO_new

    cdef double A,Bnp,Bnpa,Q,Heven,Hodd,H_val

    cdef double xi,omega,omega_circ

    cdef double dAdr,dBnpdr,dBnpadr,dxidr,dQdr,dQdprst,dHodddr,dBpdr
    cdef double aux_derivs[9]
    aux_derivs[:] = H.auxderivs(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )

    dAdr = aux_derivs[0]
    dBnpdr = aux_derivs[1]
    dBnpadr = aux_derivs[2]
    dxidr = aux_derivs[3]
    dQdr = aux_derivs[4]
    dQdprst = aux_derivs[5]
    dHodddr = aux_derivs[6]
    dBpdr = aux_derivs[7]
    dHevendr = aux_derivs[8]

    H_val, xi, A, Bnp, Bnpa, Q, Heven, Hodd, Bp = H(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )

    omega = H.omega(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )

    if omega < 0.9*omega_start:
      print(f"problem PA dynamics is extrapolating: omega ={omega}< 0.9*omega_start {omega_start}")

    params.dynamics.p_circ[1] = pphi
    omega_circ = H.omega(
        q, params.dynamics.p_circ,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )
    params.p_params.update_spins(chi1_LN, chi2_LN)

    params.p_params.omega_circ = omega_circ
    params.p_params.omega = omega
    params.p_params.H_val = H_val

    cdef (double,double) flux = RR.RR(q, p, omega, omega_circ, H_val, params)

    cdef double result = (
        xi / (2 * (1 + Bnp)) * (
            flux[1] / dpphi_dr * 2 * nu * H_val * Heven / A - xi * dQdprst
        )
    )

    return result


@cython.profile(True)
@cython.linetrace(True)
cpdef compute_pr(
    double[:] r,
    double[:] pr,
    double[:] pphi,
    double[:] omega,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    dict splines,
    double m_1,
    double m_2,
    double omega_start,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    str postadiabatic_type="analytic",
    EOBParams params=None,
):
    """
    Compute prstar at a given postadiabatic order by solving the equation
    dpphi_dr * dH_dpr * csi - flux[1] = 0, where flux[1] is the azimuthal component
    of the radiation reaction force and csi is the tortoise coordinate transformation
    factor, for prstar. This can be done either by root finding (postadiabatic_type="numeric")
    or by finding an analytic expression for the root using the potentials of the Hamiltonian
    (postadiabatic_type="analytic").

    Args:
        r (double[:]): Radial grid on which to compute prstar.
        pr (double[:]): Prstar values of the previous postadiabatic order.
        pphi (double[:]): Pphi values of the previous postadiabatic order.
        omega (double[:]): Orbital frequency values of the previous postadiabatic order.
        H (Hamiltonian_v5PHM_C): Hamiltonian.
        RR (RadiationReactionForce): Radiation Reaction Force.
        splines (dict): Dictionary containing the interpolated spin-precessing evolution.
        m_1 (double): Mass of the primary.
        m_2 (double): Mass of the secondary.
        omega_start (double): Initial orbital frequency.
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.
        tol (double): Tolerance for the root finding routine in case postadiabatic_type="numeric" is used.
        postadiabatic_type (str): If "analytic" find root analytically, otherwise numericall by a root finding method.
        params (EOBParams): Container with useful EOB parameters.

    Returns:
        (np.array) Value of prstar at a given postadiabatic order for the whole radial grid.
    """

    cdef double[:] dpphi_dr = spline_derivative(r, pphi)


    cdef int i

    tmp = splines["everything"](omega)
    chi1_v = tmp[:,4:7]
    chi2_v = tmp[:,7:10]

    cdef double[:] chi1_LN = tmp[:,0]
    cdef double[:] chi2_LN = tmp[:,1]

    cdef double[:] chi1_L = tmp[:,2]
    cdef double[:] chi2_L = tmp[:,3]

    for i in range(r.shape[0]):
        if np.abs(pr[i])<1e-14:
            x0 = 0.0
            x1 = -3e-2
        else:
            x0 = pr[i]*0.7
            x1 = pr[i]*1.3

        if postadiabatic_type == "analytic":
            pr_val = pr_eqn_analytic(
                pr[i], r[i], pphi[i], dpphi_dr[i],
                H, RR,
                chi1_v[i], chi2_v[i],
                chi1_LN[i], chi2_LN[i],
                chi1_L[i], chi2_L[i],
                m_1, m_2,
                omega_start,
                params,
                q, p
            )
        else:
            pr_solution = optimize.root(
                pr_eqn,
                pr[i],
                args=(
                    r[i], pphi[i], dpphi_dr[i],
                    H, RR, chi1_v[i], chi2_v[i],
                    chi1_LN[i],chi2_LN[i],
                    chi1_L[i],chi2_L[i],
                    m_1, m_2,
                    omega_start,
                    params,q,p
                ),
                tol=tol,

            )
            pr_val = pr_solution.x
        pr[i] = pr_val

        omega[i] = params.p_params.omega

        if omega[i] < 0.9*omega_start:
          print(f"problem PA dynamics is extrapolating: omega ={omega[i]}< 0.9*omega_start {omega_start}")


    return pr, omega

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pphi_eqn(
    pphi_sol,
    double r,
    double pr,
    double dpr_dr,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    chi1_v,
    chi2_v,
    double chi1_LN,
    double chi2_LN,
    double chi1_L,
    double chi2_L,
    double m_1,
    double m_2,
    double omega_start,
    EOBParams params,
    double[::1] q,
    double[::1] p
):
    """
    Equation dpr_dr * dH_dpr + dH_dr - flux[0] / csi, where flux[0] is the radial
    component of the radiation reaction force and csi is the tortoise coordinate
    transformation factor, to solve for ppphi in the numerical postadiabatic routine.

    Args:
        pphi_sol: Solution for pphi from the previous post-adiabatic order
        r (double): Orbital separation
        pr (double): Tortoise radial component of the momentum (prstar)
        dpr_dr (double): Radial derivative of prstar
        H (Hamiltonian_v5PHM_C): Hamiltonian
        RR (RadiationReactionForce): Radiation Reaction Force
        chi1_v : Primary dimensionless spin vector
        chi2_v : Secondary dimensionless spin vector
        chi1_LN (double): Projection of the primary dimensionless spin onto LNhat
        chi2_LN (double): Projection of the secondary dimensionless spin onto LNhat
        chi1_L (double): Projection of the primary dimensionless spin onto Lhat
        chi2_L (double): Projection of the secondary dimensionless spin onto Lhat
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        omega_start (double): Initial orbital frequency
        params (EOBParams): Container with useful EOB parameters
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.

    Returns:
        (double) Value of pphi
    """
    cdef double dH_dr,dH_dpr,H_val,csi,omega,omega_circ,result
    q[0] = r
    p[0] = pr
    p[1] =  pphi_sol[0]

    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double ap = chi1_LN * X1 + chi2_LN * X2
    cdef double am = chi1_LN * X1 - chi2_LN * X2

    cdef double dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)

    H.calibration_coeffs['dSO'] = dSO_new
    cdef double dynamics[6]

    dynamics[:] = H.dynamics(q,p, chi1_v, chi2_v, m_1, m_2,chi1_LN,chi2_LN,chi1_L,chi2_L)
    dH_dr = dynamics[0]
    dH_dpr = dynamics[2]

    omega = dynamics[3]
    H_val = dynamics[4]
    csi = dynamics[5]

    params.dynamics.p_circ[1] =  pphi_sol[0]
    omega_circ = H.omega(q,params.dynamics.p_circ,chi1_v,chi2_v,m_1,m_2,chi1_LN, chi2_LN, chi1_L, chi2_L)

    if omega < 0.9*omega_start:
      print(f"problem PA dynamics is extrapolating: omega ={omega}< 0.9*omega_start {omega_start}")

    params.p_params.update_spins(chi1_LN, chi2_LN)

    cdef (double,double) flux = RR.RR(q, p, omega,omega_circ,H_val,params)

    params.p_params.omega_circ = omega_circ
    params.p_params.omega = omega
    params.p_params.H_val = H_val

    result = dpr_dr * dH_dpr + dH_dr - flux[0] / csi

    return result



@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef double pphi_eqn_analytic(
    double pphi_sol,
    double r,
    double pr,
    double dpr_dr,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    chi1_v,
    chi2_v,
    double chi1_LN,
    double chi2_LN,
    double chi1_L,
    double chi2_L,
    double m_1,
    double m_2,
    double omega_start,
    EOBParams params,
    double[::1] q,
    double[::1] p
  ):
    """
    Equation dpr_dr * dH_dpr + dH_dr - flux[0] / csi, where flux[0] is the radial
    component of the radiation reaction force and csi is the tortoise coordinate
    transformation factor, to solve for pphi in the analytic postadiabatic routine.
    This function provides the same result as the function "pphi_eqn", but instead of
    using a root finding routine it computes the roots analytically using the
    potentials of the Hamiltonian.

    Args:
        pphi_sol: Solution for pphi from the previous post-adiabatic order
        r (double): Orbital separation
        pr (double): Tortoise radial component of the momentum (prstar)
        dpr_dr (double): Radial derivative of prstar
        H (Hamiltonian_v5PHM_C): Hamiltonian
        RR (RadiationReactionForce): Radiation Reaction Force
        chi1_v : Primary dimensionless spin vector
        chi2_v : Secondary dimensionless spin vector
        chi1_LN (double): Projection of the primary dimensionless spin onto LNhat
        chi2_LN (double): Projection of the secondary dimensionless spin onto LNhat
        chi1_L (double): Projection of the primary dimensionless spin onto Lhat
        chi2_L (double): Projection of the secondary dimensionless spin onto Lhat
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        omega_start (double): Initial orbital frequency
        params (EOBParams): Container with useful EOB parameters
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.

    Returns:
        (double) Value of pphi
    """
    q[0] = r
    p[0] = pr
    p[1] =  pphi_sol
    cdef double nu = params.p_params.nu
    params.dynamics.p_circ[1] = pphi_sol

    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double ap = chi1_LN * X1 + chi2_LN * X2
    cdef double am = chi1_LN * X1 - chi2_LN * X2

    cdef double dSO_new = dSO_poly_fit(nu, ap, am)
    H.calibration_coeffs['dSO'] = dSO_new

    cdef double dAdr,dBnpdr,dBnpadr,dxidr,dQdr,dQdprst,dHOdddr,dBpdr,xi,omega,omega_circ,result
    cdef double A,Bp,Bnp,Bnpa,Q,Heven,Hodd,H_val
    cdef double aux_derivs[9]

    aux_derivs[:] = H.auxderivs(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )


    dAdr = aux_derivs[0]
    dBnpdr = aux_derivs[1]
    dBnpadr = aux_derivs[2]
    dxidr = aux_derivs[3]
    dQdr = aux_derivs[4]
    dQdprst = aux_derivs[5]
    dHodddr = aux_derivs[6]
    dBpdr = aux_derivs[7]
    dHevendr = aux_derivs[8]

    H_val, xi, A, Bnp, Bnpa, Q, Heven, Hodd, Bp = H(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )

    omega = H.omega(
        q, p,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )

    if omega < 0.9*omega_start:
      print(f"problem PA dynamics is extrapolating: omega ={omega}< 0.9*omega_start {omega_start}")

    params.dynamics.p_circ[1] = pphi_sol
    omega_circ = H.omega(
        q, params.dynamics.p_circ,
        chi1_v, chi2_v,
        m_1, m_2,
        chi1_LN, chi2_LN,
        chi1_L, chi2_L,
    )


    params.p_params.omega_circ = omega_circ
    params.p_params.omega = omega
    params.p_params.H_val = H_val

    cdef double drdt = A/(2*nu*H_val*Heven)*(2*pr/xi*(1+Bnp)+xi*dQdprst)

    params.p_params.update_spins(chi1_LN, chi2_LN)

    cdef (double,double) flux = RR.RR(
        q, p,
        omega, omega_circ,
        H_val,
        params,
    )

    cdef double pr2 = pr*pr
    cdef double xi2 = xi*xi
    cdef double xi3 = xi2*xi
    cdef double r2 = r*r
    cdef double r3 = r2*r

    cdef double lap = chi1_L * X1 + chi2_L * X2
    cdef double lap2 = lap*lap


    cdef double C_2 = dAdr*(Bp/r2+lap2/r2*Bnpa)+A*(-2*Bp/r3-2*lap2/r3*Bnpa+lap2/r2*dBnpadr+1/r2*dBpdr)
    cdef double C_1 = dHodddr/pphi_sol*2*Heven
    cdef double C_0_pt1 = dAdr*(1+pr2/xi2*(1+Bnp) +Q) + A*(dQdr-2*pr2/xi3*dxidr*(1+Bnp)+pr2/xi2*(dBnpdr))
    cdef double tmp = 2*Heven*nu*H_val/xi
    cdef double C_0_pt2 = dpr_dr*drdt*tmp - pr/pphi_sol*tmp*flux[1]
    cdef double C_0 = C_0_pt1+C_0_pt2

    D = C_1*C_1 - 4*C_2*C_0

    if D < 0 and fabs(D) < 1e-10:
        D = 0.0
    if D < 0:
        raise ValueError

    result = (-C_1 - sqrt(D)) / (2 * C_2)

    return result


@cython.profile(True)
@cython.linetrace(True)
cpdef compute_pphi(
    double[:] r,
    double[:] pr,
    double[:] pphi,
    double[:] omega,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    splines,
    double m_1,
    double m_2,
    double omega_start,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    EOBParams params=None,
    str postadiabatic_type="analytic",
):
    """
    Compute pphi at a given postadiabatic order by solving the equation
    dpr_dr * dH_dpr + dH_dr - flux[0] / csi, where flux[0] is the radial
    component of the radiation reaction force and csi is the tortoise coordinate
    transformation factor, for pphi. This can be done either by root finding
    (postadiabatic_type="numeric") or by finding an analytic expression for the
    root using the potentials of the Hamiltonian (postadiabatic_type="analytic").

    Args:
        r (double[:]): Radial grid on which to compute prstar.
        pr (double[:]): Prstar values of the previous postadiabatic order.
        pphi (double[:]): Pphi values of the previous postadiabatic order.
        omega (double[:]): Orbital frequency values of the previous postadiabatic order.
        H (Hamiltonian_v5PHM_C): Hamiltonian.
        RR (RadiationReactionForce): Radiation Reaction Force.
        splines (dict): Dictionary containing the interpolated spin-precessing evolution.
        m_1 (double): Mass of the primary.
        m_2 (double): Mass of the secondary.
        omega_start (double): Initial orbital frequency.
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.
        tol (double): Tolerance for the root finding routine in case postadiabatic_type="numeric" is used.
        params (EOBParams): Container with useful EOB parameters.
        postadiabatic_type (str): If "analytic" find root analytically, otherwise numericall by a root finding method.

    Returns:
        (np.array) Value of pphi at a given postadiabatic order for the whole radial grid.
    """

    cdef double[:] dpr_dr = spline_derivative(r, pr)

    cdef int i
    tmp = splines["everything"](omega)
    chi1_v = tmp[:,4:7]
    chi2_v = tmp[:,7:10]

    cdef double[:] chi1_LN = tmp[:,0]
    cdef double[:] chi2_LN = tmp[:,1]

    cdef double[:] chi1_L = tmp[:,2]
    cdef double[:] chi2_L = tmp[:,3]

    for i in range(r.shape[0]):
        if postadiabatic_type == "analytic":
            pphi_val = pphi_eqn_analytic(
                pphi[i], r[i], pr[i], dpr_dr[i],
                H, RR,
                chi1_v[i], chi2_v[i],
                chi1_LN[i], chi2_LN[i],
                chi1_L[i], chi2_L[i],
                m_1, m_2,
                omega_start,
                params,
                q, p,
            )
        else:
            pphi_solution = optimize.root(
                pphi_eqn,
                pphi[i],
                args=(
                    r[i], pr[i], dpr_dr[i],
                    H, RR,
                    chi1_v[i], chi2_v[i],
                    chi1_LN[i], chi2_LN[i],
                    chi1_L[i], chi2_L[i],
                    m_1, m_2,
                    omega_start,
                    params,
                    q, p,
                ),
                tol=tol,
            )
            pphi_val = pphi_solution.x

        omega[i] = params.p_params.omega
        pphi[i] = pphi_val

    return pphi, omega

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_postadiabatic_solution(
    double[:] r,
    double[:] pphi,
    double[:] omega,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    splines,
    double m_1,
    double m_2,
    double[::1] q,
    double[::1] p,
    double tol=1e-12,
    int order=8,
    str postadiabatic_type="analytic",
    EOBParams params=None,
):
    """
    Compute postadiabatic solution for prstar, pphi and omega up to a certain
    postadiabatic order. This can be done either by root finding
    (postadiabatic_type="numeric") or by finding an analytic expression for the
    root using the potentials of the Hamiltonian (postadiabatic_type="analytic").

    Args:
        r (double[:]): Radial grid on which to compute postadiabatic solution.
        pphi (double[:]): Pphi values computed from the adiabatic solution.
        omega (double[:]): Orbital frequency values of the adiabatic solution.
        H (Hamiltonian_v5PHM_C): Hamiltonian.
        RR (RadiationReactionForce): Radiation Reaction Force.
        splines (dict): Dictionary containing the interpolated spin-precessing evolution.
        m_1 (double): Mass of the primary.
        m_2 (double): Mass of the secondary.
        q (double[::1]): Position q=[r,phi]= np.zeros(2). These values are updated in the root finding routine.
        p (double[::1]): Canonical momentum p=[pr,pphi]=np.zeros(2). These values are updated in the root finding routine.
        tol (double): Tolerance for the root finding routine in case postadiabatic_type="numeric" is used.
        order (int): Postadiabatic order up to which to obtain the solution for prstar and pphi.
        postadiabatic_type (str): If "analytic" find root analytically, otherwise numericall by a root finding method.
        params (EOBParams): Container with useful EOB parameters.

    Returns:
        (prstar, pphi, omega) up to a given postadiabatic order for the whole radial grid.
    """
    pr = np.zeros(r.size)
    cdef int i,n,parity
    cdef double tol_current
    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double omega_start = splines["everything"].x[0]

    for n in range(1, order+1):
        tol_current = 1e-3 / 10**n
        parity = n % 2
        if n>=7:
            tol_current=tol
        if parity:
            pr, _ = compute_pr(
                r,
                pr,
                pphi,
                omega,
                H,
                RR,
                splines,
                m_1,
                m_2,
                omega_start,
                q,
                p,
                tol=tol_current,
                postadiabatic_type=postadiabatic_type,
                params=params,
            )
        else:
            pphi, _ = compute_pphi(
                r,
                pr,
                pphi,
                omega,
                H,
                RR,
                splines,
                m_1,
                m_2,
                omega_start,
                q,
                p,
                tol=tol_current,
                postadiabatic_type=postadiabatic_type,
                params=params,
            )

        tmp = splines["everything"](omega)
        chi1_v = tmp[:,4:7]
        chi2_v = tmp[:,7:10]

        chi1_LN = tmp[:,0]
        chi2_LN = tmp[:,1]

        chi1_L = tmp[:,2]
        chi2_L = tmp[:,3]

        ap = chi1_LN * X1 + chi2_LN * X2
        am = chi1_LN * X1 - chi2_LN * X2
        dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)

        for i in range(r.size):
            q[0] = r[i]
            p[0] = pr[i]
            p[1] = pphi[i]

            params.p_params.chi1_v = chi1_v[i]
            params.p_params.chi2_v = chi2_v[i]

            params.p_params.chi_1 = chi1_LN[i]
            params.p_params.chi_2 = chi2_LN[i]

            params.p_params.chi1_L = chi1_L[i]
            params.p_params.chi2_L = chi2_L[i]

            H.calibration_coeffs['dSO'] = dSO_new[i]

            om = H.omega(
                q, p,
                chi1_v[i], chi2_v[i],
                m_1, m_2,
                chi1_LN[i], chi2_LN[i],
                chi1_L[i], chi2_L[i],
            )
            omega[i] = om

    return pr, pphi, omega

@cython.profile(True)
@cython.linetrace(True)
cpdef compute_postadiabatic_dynamics(
    double omega_ref,
    double omega_start,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    chi_1,
    chi_2,
    double m_1,
    double m_2,
    dict splines,
    t_pn: np.array,
    dynamics_pn: np.array,
    double tol=1e-12,
    EOBParams params=None,
    int order=1,
    str postadiabatic_type="analytic",
    int window_length=10,
    only_first_n=None,
):
    """
    Compute postadiabatic dynamics from the spins specified at a certain reference frequency and
    with a certain starting frequency. The postadiabatic evolution can be done
    either by root finding (postadiabatic_type="numeric") or by finding an analytic expression for
    the roots of prstar and pphi using the potentials of the Hamiltonian (postadiabatic_type="analytic").

    Args:
        omega_ref (double): Reference orbital frequency at which the spins are specified.
        omega_start (double): Initial orbital frequency.
        H (Hamiltonian_v5PHM_C): Hamiltonian.
        RR (RadiationReactionForce): Radiation Reaction Force.
        chi_1 : Primary dimensionless spin vector.
        chi_2 : Secondary dimensionless spin vector.
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        splines (dict): Dictionary containing the interpolated spin-precessing evolution.
        params (EOBParams): Container with useful EOB parameters
        tol (double): Tolerance for the root finding routine in case postadiabatic_type="numeric" is used.
        order (int): Postadiabatic order up to which to obtain the solution for prstar and pphi.
        postadiabatic_type (str): If "analytic" find root analytically, otherwise numericall by a root finding method.
        window_length (int): Minimum number of points in the radial grid.
        only_first_n (int): Compute the postadiabatic only on a specified number of points.

    Returns:
        (postadiabatic_dynamics, omega, lN_dyn, splines) where postadiabatic_dynamics includes also the
        quantities needed to compute the waveform modes = [t,r,phi,pr,pphi,H,omega,omega_circ,chi1LN,chi2LN].

    """

    cdef double Mtot = m_1 + m_2
    cdef double nu = m_1*m_2/Mtot/Mtot

    tmp = splines["everything"](omega_start)

    cdef double chi1_LN = tmp[0]
    cdef double chi2_LN = tmp[1]

    cdef double chi1_L = tmp[2]
    cdef double chi2_L = tmp[3]

    chi1_v = tmp[4:7]
    chi2_v = tmp[7:10]
    LNhat = tmp[10:13]

    params.p_params.omega = omega_start

    params.p_params.chi1_v = chi1_v
    params.p_params.chi2_v = chi2_v

    params.p_params.chi_1, params.p_params.chi_2 = chi1_LN, chi2_LN
    params.p_params.chi1_L, params.p_params.chi2_L = chi1_L, chi2_L

    params.p_params.update_spins(chi1_LN, chi2_LN)

    cdef double X1 = params.p_params.X_1
    cdef double X2 = params.p_params.X_2

    cdef double ap = chi1_LN * X1 + chi2_LN * X2
    cdef double am = chi1_LN * X1 - chi2_LN * X2
    cdef double chi_perp_eff = np.linalg.norm(chi1_v*X1+chi2_v*X2-ap*LNhat)

    cdef double dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)
    H.calibration_coeffs['dSO'] = dSO_new

    cdef double r0
    r0, _, _ =computeIC_augm(omega_start, H, RR, chi1_v, chi2_v, m_1, m_2, params=params)
    cdef double chi_eff = ap

    r_ISCO, _ = Kerr_ISCO(chi1_LN, chi2_LN, X1, X2)

    # Phenomenological prefactor to r_ISCO to adjust r_final in the precessing case
    cdef double r_final_prefactor = 2.7 + (chi_eff - 0.5*chi_perp_eff)*(1-4.*nu)

    # Compute the final point of the radial grid
    cdef double r_final = max(11.0, r_final_prefactor * r_ISCO)

    cdef double dr0 = 0.1
    cdef int r_size = int(np.ceil((r0 - r_final) / dr0))
    cdef double r_range = r0 - r_final

    # Compute now the number of precessing cycles (only in the aligned-spin limit)
    cdef double chi_in_plane = chi1_v[0]**2.+chi1_v[1]**2. + chi2_v[0]**2.+chi2_v[1]**2.
    cdef int r_size_new = 0
    cdef double precessiong_cycles = 0.0
    cdef double dr0_new = 0.1

    omega_pn = dynamics_pn[:,-1]
    lN_pn = dynamics_pn[:,:3]
    if chi_in_plane > 0 :
      precession_cycles = compute_prec_cycles(r_final,t_pn, omega_pn, lN_pn)

      # Multiply the number of precession cycles by 20, so that the have 20 points per precession cycle,
      # however as we are making a uniform radial grid below this condition will not be exactly fulfilled
      r_size_new = int(np.ceil(precession_cycles*20))


    # Some checks to decide whether the radial grid is large enough to perform the postadiabatic evolution
    if r_size <= 10 or r0<=11.5 or r_final >= r0:
        raise ValueError
    elif r_size < window_length + 2:
        r_size = window_length + 2

    # If the number of precessing cycles is zero, we use 0.1 as the default step size of the radial grid
    if r_size_new ==0:
      dr0_new = 0.1
      r_size_new = int(np.ceil(r_range/dr0_new))
    else:
      dr0_new = r_range/r_size_new

    # Put some upper and lower bounds to the step size of the radial grid
    if dr0_new < 0.05:
      dr0_new = 0.05
      r_size_new = int(np.ceil(r_range/dr0_new))

    if dr0_new > 0.1:
      dr0_new = 0.1
      r_size_new = int(np.ceil(r_range/dr0_new))

    # Compute the final grid
    r, _ = np.linspace(r0, r_final, num=r_size_new, endpoint=True, retstep=True)

    # Option only used in the PA initial conditions for the non-PA model
    if only_first_n is not None:
      r = r[:only_first_n]

    # First guess at circular omega, from Kepler's law
    omega = r**(-3./2)

    # Arrays for the position and canonical momentum
    cdef double[::1] q = np.zeros(2)
    cdef double[::1] p = np.zeros(2)

    # Compute the adiabatic solution
    pphi = compute_adiabatic_solution(r, omega, H, splines, m_1, m_2, q, p, params, tol=tol,)

    # Update the circular omega with the adiabatic solution
    for i in range(r.size):
        q[0] = r[i]
        p[1] = pphi[i]

        tmp = splines["everything"](omega[i])

        chi1_LN = tmp[0]
        chi2_LN = tmp[1]

        chi1_L = tmp[2]
        chi2_L = tmp[3]

        chi1_v = tmp[4:7]
        chi2_v = tmp[7:10]

        ap = chi1_LN * X1 + chi2_LN * X2
        am = chi1_LN * X1 - chi2_LN * X2
        dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)

        H.calibration_coeffs['dSO'] = dSO_new

        om = H.omega(
            q, p,
            chi1_v, chi2_v,
            m_1, m_2,
            chi1_LN, chi2_LN,
            chi1_L, chi2_L,
        )

        omega[i] = om
        if omega[i]< 0.9*omega_start :
          print(f"WARNING PA DYNAMICS IS EXTRAPOLATING!")


    # Compute the PA solution
    pr, pphi, omega = compute_postadiabatic_solution(
        r,
        pphi,
        omega,
        H,
        RR,
        splines,
        m_1,
        m_2,
        q,
        p,
        tol=tol,
        order=order,
        postadiabatic_type=postadiabatic_type,
        params=params,
    )

    dt_dr = np.zeros(r.shape[0])
    dphi_dr = np.zeros(r.shape[0])
    cdef double dH_dpr,dH_dpphi,csi
    cdef double dyn[6]
    cdef double[::1] p_circ = np.zeros(2)

    dyn_augm = []

    for i in range(r.shape[0]):
        q = np.array([r[i], 0])
        p = np.array([pr[i], pphi[i]])

        tmp = splines["everything"](omega[i])

        chi1_LN = tmp[0]
        chi2_LN = tmp[1]

        chi1_L = tmp[2]
        chi2_L = tmp[3]

        chi1_v = tmp[4:7]
        chi2_v = tmp[7:10]

        ap = chi1_LN * X1 + chi2_LN * X2
        am = chi1_LN * X1 - chi2_LN * X2

        dSO_new = dSO_poly_fit(params.p_params.nu, ap, am)
        H.calibration_coeffs['dSO'] = dSO_new

        dyn[:] = H.dynamics(q, p, chi1_v, chi2_v, m_1, m_2,chi1_LN, chi2_LN, chi1_L, chi2_L)
        dH_dpr = dyn[2]
        dH_dpphi = dyn[3]
        H_val = dyn[4]
        csi = dyn[5]

        dt_dr[i] = 1 / (dH_dpr * csi)
        dphi_dr[i] = dH_dpphi / (dH_dpr * csi)
        omega[i] = dH_dpphi

        p_circ[1] = p[1]
        omega_circ = H.omega(q, p_circ, chi1_v, chi2_v, m_1, m_2, chi1_LN, chi2_LN, chi1_L, chi2_L)
        dyn_augm.append([H_val,omega[i],omega_circ,chi1_LN,chi2_LN])

    # Compute integrals to obtain the time array and the orbital phase
    t = univariate_spline_integral(r,dt_dr)
    phi = univariate_spline_integral(r,dphi_dr)

    postadiabatic_dynamics = np.c_[t, r, phi, pr, pphi, dyn_augm]

    return postadiabatic_dynamics, omega



@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
cpdef compute_combined_dynamics_exp_v1(
    double omega_ref,
    double omega_start,
    Hamiltonian_v5PHM_C H,
    RadiationReactionForce RR,
    double m_1,
    double m_2,
    chi_1,
    chi_2,
    double tol=1e-12,
    EOBParams params=None,
    double step_back=50,
    str backend="ode",
    int PA_order=8,
    str postadiabatic_type="analytic",
):

    """
    Compute the dynamics by using a postadiabatic scheme up to a certain separation from which an
    ODE integration is started. The postadiabatic evolution can be done either by root finding
    (postadiabatic_type="numeric") or by finding an analytic expression for the roots of prstar
    and pphi using the potentials of the Hamiltonian (postadiabatic_type="analytic").

    Args:
        omega_ref (double): Reference orbital frequency at which the spins are specified.
        omega_start (double): Initial orbital frequency.
        H (Hamiltonian_v5PHM_C): Hamiltonian.
        RR (RadiationReactionForce): Radiation Reaction Force.
        m_1 (double): Mass of the primary
        m_2 (double): Mass of the secondary
        chi_1 : Primary dimensionless spin vector.
        chi_2 : Secondary dimensionless spin vector.
        tol (double): Tolerance for the root finding routine in case postadiabatic_type="numeric" is used.
        params (EOBParams): Container with useful EOB parameters
        step_back (float, optional): Amount of time to step back for fine interpolation. Defaults to 250.
        PA_order (int): Postadiabatic order up to which to obtain the solution for prstar and pphi.
        postadiabatic_type (str): If "analytic" find root analytically, otherwise numericall by a root finding method.

        Returns:
            (tuple): Low and high sampling rate dynamics, unit Newtonian orbital angular momentum, assembled dynamics
                     and the index splitting the low and high sampling rate dynamics.

    """
    # Save initial frequency to use it afterwards for the roll-off
    cdef double omega_start_0 = omega_start

    combined_t, combined_y = compute_quasiprecessing_PNdynamics_opt(
        omega_ref,
        0.9*omega_start,
        m_1, m_2,
        chi_1, chi_2,
    )

    omegaPN_f = combined_y[:, -1][-1]

    splines = build_splines_PN(
        combined_t,
        combined_y,
        m_1, m_2,
        omega_start,
    )

    try:

        postadiabatic_dynamics, omega_pa = compute_postadiabatic_dynamics(
            omega_ref,
            omega_start,
            H,
            RR,
            chi_1,
            chi_2,
            m_1,
            m_2,
            splines,
            combined_t,
            combined_y,
            tol=tol,
            params=params,
            order=PA_order,
            postadiabatic_type=postadiabatic_type,
            window_length=10,
        )

        PA_success = True
        e_id = -1
        ode_y_init = postadiabatic_dynamics[e_id, 1:5]

    except ValueError as e:
        PA_success = False
        ode_y_init = None
        omega_pa = [omega_start]

    # Set omega_start to the last frequency of the PA evolution
    omega_start = omega_pa[-1]

    (
        ode_dynamics_low,
        ode_dynamics_high,
        dynamics,
        idx_restart
    ) = compute_dynamics_prec_opt(
        omega_ref,
        omega_start,
        omegaPN_f,
        H,
        RR,
        m_1,
        m_2,
        splines,
        combined_t,
        combined_y,
        params,
        rtol=1e-8,
        atol=1e-8,
        step_back=step_back,
        y_init=ode_y_init,
    )


    if PA_success is True:

        # Interpolate the PA dynamics in its full range. Note that for the AS model this is only
        # done in a window before the ODE integration to smooth the timestep transition. For the
        # precessing model this is done on the full range of the post-adiabatic dynamics to ensure
        # enough resolution to accurately describe the rotations from the co-precessing frame to the
        # inertial frame

        t_pa = postadiabatic_dynamics[:, 0]

        ode_dynamics_low[:, 0] += t_pa[e_id]
        ode_dynamics_high[:, 0] += t_pa[e_id]

        # This is the part of the PA dynamics which is smoothed in
        # timestep until it reaches the one obtained solving ODEs
        t_ode_low = ode_dynamics_low[:, 0]

        dt_pa_first = t_pa[1] - t_pa[0]
        dt_ode_init = t_ode_low[1] - t_ode_low[0]

        # Estimate initial delta_t from the starting orbital frequency (this should correspond to the maximum timestep in the PA dynamics)
        dt0 = 2.*np.pi/omega_start_0

        while dt_pa_first>500:
            if  dt_pa_first > dt0:
              dt_pa_first = dt0/5.
            else:
              dt_pa_first = dt0/10.
            dt0 = dt_pa_first

        t_pa_first = t_pa[0]
        t_ode_init = t_ode_low[0]

        dt = dt_ode_init
        t = t_ode_init - dt
        t_new = []
        step_multiplier = 1.3

        while True:
            t_new.append(t)

            if step_multiplier * dt < dt_pa_first:
                dt *= step_multiplier

            t -= dt

            if t < t_pa_first:
                break

        # Add the initial time
        t_new.append(t_pa[0])

        # Reverse the timeseries
        t_new = t_new[::-1]

        # Interpolate window dynamics except the spins projections (otherwise bad things may happen due to the large timesteps of the PA dynamics)
        window_dynamics_interp = CubicSpline(t_pa, np.c_[postadiabatic_dynamics[:,:-2], omega_pa])
        tmp_window = window_dynamics_interp(t_new)

        window_dynamics = tmp_window[:,:-1]
        omega_new = tmp_window[:,-1]

        # Compute the spin projections onto LNhat from the new interpolated orbital frequency
        tmp = splines["everything"](omega_new)
        chi1_LN_window = tmp[:,0]
        chi2_LN_window = tmp[:,1]
        postadiabatic_dynamics_v1 = np.c_[window_dynamics, chi1_LN_window,chi2_LN_window]

        combined_dynamics = np.vstack((postadiabatic_dynamics_v1[:e_id], ode_dynamics_low))

    else:
        combined_dynamics = ode_dynamics_low

    dynamics = np.vstack((combined_dynamics, ode_dynamics_high))

    return combined_dynamics, ode_dynamics_high,combined_t,combined_y,splines,dynamics



@cython.boundscheck(True)
@cython.profile(True)
@cython.linetrace(True)
@cython.cdivision(True)
cpdef double compute_prec_cycles(r_final: float,t_pn: np.array, omega_pn: np.array, lN_pn:np.array):
    """
    Estimate the number of precession cycles from LNhat, computed as the phase of the precession
    frequency at the final point of the radial grid divided by 2 pi.

    Args:
        r_final (float): Last point of the radial grid.
        t_pn (np.array): Time array of the PN evolution.
        omega_pn (np.array): Orbital frequency from the PN evolution.
        lN_pn (np.array): LNhat obtained from the PN evolution.

    Returns:
        (double) Number of precession cycles.
    """

    # Compute the LNhat quaternion
    lN_quat = quaternion.as_quat_array(np.c_[np.zeros(len(lN_pn)), lN_pn])


    # Compute the angular velocity
    omega_prec_arr = quaternion.angular_velocity(lN_quat,t_pn)/2.0
    om_prec_norm = np.sqrt(np.einsum("ij,ij->i", omega_prec_arr, omega_prec_arr))

    # Compute the value of the precession frequency at r_final
    r_pn = omega_pn**(-3./2.)
    iut = CubicSpline(1./r_pn,t_pn)
    iom_prec = CubicSpline(t_pn,om_prec_norm)

    # Compute the phase
    ph_om_prec = iom_prec.antiderivative()(t_pn)
    iph_prec = CubicSpline(t_pn,ph_om_prec)


    # Evaluate the phase at t_final
    t_final = iut(1./r_final)
    prec_cycles = iph_prec(t_final)/(2.*np.pi)


    return prec_cycles



@cython.boundscheck(False)
@cython.nonecheck(False)
@cython.initializedcheck(False)
@cython.profile(True)
@cython.linetrace(True)
@cython.cdivision(True)
def spline_derivative(x: np.array, y: np.array)->np.array:
      """
      Compute derivative of y(x) using cubic splines.

      Args:
          x (np.array): Dependent variable.
          y (np.array): Independent variable.

      Returns:
          (np.array) derivaive dy/dx. Note that the order is reversed as x will be
                     the orbital separation which is monotonically decreasing, while
                     the interpolation requires x to be monotonically increasing.
      """
      intrp = CubicSpline(x[::-1], y[::-1])
      deriv = intrp.derivative()(x[::-1])[::-1]
      return deriv



#!/usr/bin/env python3
import logging
from abc import ABC
from typing import Any, Dict, Union

import lal
import numpy as np
from pyseobnr.models.model import Model
from rich.logging import RichHandler
from rich.traceback import install
from scipy.interpolate import CubicSpline

from .default_settings import default_unfaitfulness_mode_by_mode_settings
from .unfaithfulness_mode_by_mode import (
    condition_pycbc_series,
    fast_unfaithfulness_mode_by_mode,
    generate_dominant_mode_pol_LAL,
    get_padded_length,
)

try:
    from waveform_tools.mismatch.auxillary_funcs import *

except ImportError:
    print("waveform_tools is not installed, some metrics not available")


from pycbc.types import FrequencySeries, TimeSeries
from pycbc.waveform import fd_approximants

# Setup the logger to work with rich
logger = logging.getLogger(__name__)
logger.addHandler(RichHandler(rich_tracebacks=True, markup=True))
logger.setLevel("INFO")
# Setup rich to get nice tracebacks
install()


class Metric(ABC):
    def __init__(self):
        pass

    def __call__(self):
        pass


class UnfaithfulnessModeByModeLAL(Metric):
    def __init__(self, settings: Dict[Any, Any] = None):
        self.settings = default_unfaitfulness_mode_by_mode_settings()
        if settings is not None:
            self.settings.update(**settings)

        self.debug = self.settings.get("debug", False)
        self.masses = self.settings.get(
            "masses", np.array([10.0, 50.0, 100.0, 150.0, 200.0])
        )
        self.buffer = self.settings.get("buffer", 1.35)
        self.fmax = self.settings.get("fmax", 2048.0)
        self.name = "UnfaithfulnessModeByModeLAL"

    def __call__(
        self,
        target: Union[str, Model],
        model: Union[str, Model],
        params: Dict[Any, Any] = None,
        ell: int = 2,
        m: int = 2,
    ) -> float:
        return self._evaluate(target, model, params=params, ell=ell, m=m)

    def _evaluate(
        self,
        target: Union[str, Model],
        model: Union[str, Model],
        params: Dict[Any, Any] = None,
        ell: int = 2,
        m: int = 2,
        M=50.0,
    ) -> float:

        # Ensure that the input parameters of everrything are consistent

        logger.debug("Sanity checking inputs")
        # self._sanity_check_inputs(target, model, params)
        logger.debug("Generating waveforms")
        if isinstance(target, Model):
            # This is one of fully wrapped models, don't need to generate things, just grab them
            h1 = TimeSeries(
                np.real(target.waveform_modes[f"{ell},{m}"]), delta_t=target.delta_T
            )
            t1 = target.t
            #omega1 = target.omega0
            # Estimate the actual starting frequency from (2,2) mode
            intrp = CubicSpline(target.t,np.unwrap(np.angle(target.waveform_modes['2,2'])))
            omega1 = np.abs(intrp.derivative()(target.t[0]))/2.
            logger.debug("Target was a Model, grabbed stuff")
        else:
            # Assume that otherwise it's a call to LAL and special care has to be taken
            if params is not None:
                params_target = self._wrap_params(params, target)
            else:
                raise ValueError(
                    "The target specified is not a full Model. You must supply params!"
                )

            h1 = generate_dominant_mode_pol_LAL(params_target, f_max=self.fmax)
            omega1 = params["omega0"]  # By construction

        if isinstance(model, Model):
            # This is one of fully wrapped models, don't need to generate things, just grab them
            h2 = TimeSeries(
                np.real(model.waveform_modes[f"{ell},{m}"]), delta_t=model.delta_T
            )
            t2 = model.t
            omega2 = model.omega0
        else:
            # Assume that otherwise it's a call to LAL and special care has to be taken
            if params is not None:
                params_model = self._wrap_params(params, model)

            else:
                raise ValueError(
                    "The model specified is not a full Model. You must supply params!"
                )

            # Ugly logic to ensure consistency of Fourier-domain waveforms
            if params_model.domain == "FD" and isinstance(h1, TimeSeries):
                # We want to generate and FD waveform that will match the given TD
                # waveform when it's FFTed.
                padsize = get_padded_length(h1)
                padded_duration = h1.delta_t * padsize * M * lal.MTSUN_SI
                params_model.delta_f = 1 / padded_duration
                nyquist_f = 0.5 / (h1.delta_t * M * lal.MTSUN_SI)
                f_max = nyquist_f

            else:
                f_max = self.fmax

            # print(params_model)
            h2 = generate_dominant_mode_pol_LAL(params_model, f_max=f_max)
            omega2 = params["omega0"]  # by construction!

        omega0 = np.max((omega1, omega2))
        flow = self.buffer * m * omega0 / (2 * np.pi)

        # Some convoluted logic for conditioning
        # Situation 1: both inputs are in time domain
        # Figure out the length of the longer waveform
        # We use this to pad the waveforms
        # Situtation 2: one of the inputs is in frequency domain
        # Figure out what we do.
        if isinstance(h1, TimeSeries) and isinstance(h2, TimeSeries):
            N = np.max((len(h1), len(h2)))
            h1 = condition_pycbc_series(h1, n=N)
            h2 = condition_pycbc_series(h2, n=N)
        elif isinstance(h1, TimeSeries) and isinstance(h2, FrequencySeries):
            h1 = condition_pycbc_series(h1)
        elif isinstance(h1, FrequencySeries) and isinstance(h2, TimeSeries):
            h2 = condition_pycbc_series(h2)

        matches = fast_unfaithfulness_mode_by_mode(h1, h2, flow, Ms=self.masses, psd_t = 'aLIGO')
        # Get the highest unfaithulness
        mm = 1 - np.min(matches)
        if self.debug:
            return 1 - matches
        return mm

    def _sanity_check_inputs(self, target, model, params):

        if (
            isinstance(target, Model) and isinstance(model, Model)
        ) and params is not None:
            logger.warning("Ignoring the params dict, as it was not needed!")
            return
        # At this point only one (target,model) can be a Model
        if isinstance(target, Model) and params is not None:
            q_t = target.q
            chi1_t = target.chi1
            chi2_t = target.chi2
            q = params["q"]
            chi1 = params["chi1"]
            chi2 = params["chi1"]
            assert np.allclose(q, q_t)
            # assert np.allclose(chi1_t, chi1)
            # assert np.allclose(chi2_t, chi2)
        elif isinstance(model, Model) and params is not None:
            q_t = model.q
            chi1_t = model.chi1
            chi2_t = model.chi2
            q = params["q"]
            chi1 = params["chi1"]
            chi2 = params["chi1"]
            assert np.allclose(q, q_t)
            # assert np.allclose(chi1_t[-1], chi1)
            # assert np.allclose(chi2_t, chi2)

    def _wrap_params(self, params, approx, M=50.0):
        q = params["q"]
        chi1 = params["chi1"]
        chi2 = params["chi2"]
        omega0 = params["omega0"]
        dt = params["dt"] * M * lal.MTSUN_SI
        df = params["df"]
        domain = self._get_domain(approx)
        if domain == "TD":
            df = None
        else:
            dt = None

        m1 = q / (1 + q)
        m2 = 1 - m1
        f_ref = f_min = omega0 / (M * np.pi * lal.MTSUN_SI)
        wf_pars = waveform_params(
            m1=M * m1,
            m2=M * m2,
            s1x=chi1[0],
            s1y=chi1[1],
            s1z=chi1[2],
            s2x=chi2[0],
            s2y=chi2[1],
            s2z=chi2[2],
            iota=0.0,
            phi=0.0,
            f_ref=f_ref,
            f_min=f_min,
            distance=1.2342710325965468e25,
            delta_t=dt,
            delta_f=df,
            wf_param=None,
            approx=approx,
            domain=domain,
            ecc=0.0,
            mean_anomaly=0.0,
        )
        return wf_pars

    def _get_domain(self, approx):
        fd_approxs = fd_approximants()
        if approx not in fd_approxs or approx == "SEOBNRv4":
            return "TD"
        else:
            return "FD"



"""
Contains functions associated with waveform construction, mostly for merger-ringdown.
"""

from copy import deepcopy
from typing import Any, Dict

import numpy as jnp
from lalinference.imrtgr import nrutils
from pygsl_lite import spline
from pyseobnr.auxiliary.mode_mixing.auxiliary_functions_modemixing import *
from scipy.interpolate import CubicSpline, InterpolatedUnivariateSpline
from scipy.optimize import root_scalar
from scipy.signal import argrelmax, argrelmin

from ..fits.EOB_fits import *
from ..fits.IV_fits import InputValueFits
from ..fits.MR_fits import MergerRingdownFits
from ..waveform.waveform import *
from .compute_MR import compute_MR_mode_free
from ..utils.utils_precession_opt import compute_omegalm_P_frame


def concatenate_modes(hlms_1: Dict[Any, Any], hlms_2: Dict[Any, Any]) -> Dict[Any, Any]:
    """Concatenate 2 dictionaries of waveform modes

    This is used to put together the low and fine sampling waveform modes.

    Note:
        Assumes that the 2 dictionaries have the same keys.

    Args:
        hlms_1 (Dict[Any,Any]): First dictionary of waveform modes to concatenate
        hlms_2 (Dict[Any,Any]): Second dictionary of waveform modes to concatenate

    Returns:
        Dict[Any,Any]: Concatenated modes
    """
    hlms = {}
    for key in hlms_1.keys():
        hlms[key] = np.concatenate((hlms_1[key], hlms_2[key]))
    return hlms


def interpolate_modes_fast(
    t_old: np.ndarray,
    t_new: np.ndarray,
    modes_dict: Dict[Any, Any],
    phi_orb: np.ndarray,
    m_max: int = 5,
) -> Dict[str, Any]:
    """Construct inertial frame modes on a new regularly
    spaced time grid.

    Does this by employing a carrier
    signal, see the idea in [Cotesta2020]_ .

    Uses a custom version of CubicSpline that is faster, but
    cannot handle derivatives or integrals.

    Args:
        t_old (np.ndarray): Original unequally spaced time array
        t_new (np.ndarray): New equally spaced time array
        modes_dict (dict): Dictionary containing *complex* modes
        phi_orb (np.ndarray): Orbital phase
        m_max (int): Max m appearing in the modes

    Returns:
        dict: Dictionary of modes interpolated onto t_new
    """
    modes_intrp = {}

    n = len(t_old)
    intrp_orb = spline.cspline(n)
    intrp_orb.init(t_old, phi_orb)
    phi_orb_interp = intrp_orb.eval_e_vector(t_new)
    tmp_store = np.zeros(len(phi_orb_interp), dtype=np.complex128)
    intrp_re = spline.cspline(n)
    intrp_im = spline.cspline(n)

    factors = np.zeros((m_max, len(phi_orb_interp)), dtype=np.complex128)
    compute_factors(phi_orb_interp, m_max, factors)
    for key, item in modes_dict.items():
        m = key[1]

        tmp = item * np.exp(1j * m * phi_orb)

        intrp_re.init(t_old, tmp.real)
        intrp_im.init(t_old, tmp.imag)
        result_re = intrp_re.eval_e_vector(t_new)
        result_im = intrp_im.eval_e_vector(t_new)
        unrotate_leading_pn(result_re, result_im, factors[m - 1], tmp_store)
        modes_intrp[key] = 1 * tmp_store
    return modes_intrp


def compute_IMR_modes(
    t,
    hlms,
    t_for_compute,
    hlms_for_compute,
    m1,
    m2,
    chi1,
    chi2,
    t_attach,
    f_nyquist,
    lmax_nyquist,
    mixed_modes=[(3, 2), (4, 3)],
    final_state=None,
    qnm_rotation=0.0,
):
    """This computes the IMR modes given the inspiral modes and the
    attachment time.

    Args:
        t (np.ndarray): The interpolated time array of the inspiral modes
        hlms (np.ndarray): Dictionary containing the inspiral modes
        t_for_compute (np.ndarray): The fine dynamics time array
        hlms_for_compute (np.ndarray): The waveform modes on the fine dynamics
        m1 (float): Mass of primary
        m2 (float): Mass of secondary
        chi1 (float): z-component of the primary dimensionless spin
        chi2 (float): z-component of the secondary dimensionless spin
        t_attach (float): Attachment time
        f_nyquist (float): Nyquist frequency, needed for checking that RD frequency is resolved
        lmax_nyquist (int): Determines for which modes the nyquist test is applied for
        mixed_modes (List): List of mixed modes to consider. Defaults to [(3,2),(4,3)]
        final_state (List): Final mass and spin of the remnant. Default to None. If None,
                            compute internally.
        qnm_rotation (float): Factor rotating the QNM mode frequency in the co-precessing frame (Eq. 33 of Hamilton et al.)

    Returns:
        dict: Dictionary containing the waveform modes
    """

    # We want to attach the ringdown always at the same time,
    # regardless of the sampling rate, i.e. all the functions
    # are evaluated at the true attachment time, not just the
    # closest grid point, as was done in v4.
    # This requires one to be somewhat careful in the construction

    # Dictionary that will hold the final modes
    hIMR = {}

    # First find the closest point on the time grid which is
    # *before* the attachment time. We do this twice,
    # because for the (5,5) mode the attachment time is
    # different from other modes

    # All modes except (5,5)
    idx = jnp.argmin(jnp.abs(t - t_attach))
    if t[idx] > t_attach:
        idx -= 1

    # Time at the grid-point just before the attachment point
    t_match = t[idx]

    # (5,5) mode
    idx_55 = jnp.argmin(jnp.abs(t - (t_attach - 10)))
    if t[idx_55] > t_attach - 10:
        idx_55 -= 1
    t_match_55 = t[idx_55]

    # The time spacing. This assumes that we have already
    # interpolated the modes to equal spacing
    dt = np.diff(t)[0]
    N = int(10 / dt) + 1

    # Figure out the duration of the ringdown. Taken to be 30
    # times the damping time of the (2,2) mode
    # To compute QNM, get the final mass and spin

    if final_state:
        final_mass, final_spin = final_state
    else:

        final_mass = nrutils.bbh_final_mass_non_precessing_UIB2016(m1, m2, chi1, chi2)
        final_spin = nrutils.bbh_final_spin_non_precessing_HBR2016(
            m1, m2, chi1, chi2, version="M3J4"
        )

    omega_complex = compute_QNM(2, 2, 0, final_spin, final_mass).conjugate()

    # For non-precessing system this does not make any difference
    omega_complex = compute_omegalm_P_frame(omega_complex, 2, qnm_rotation)
    damping_time = 1 / np.imag(omega_complex)
    # The length of the ringdown rounded to closest M
    ringdown_time = int(30 * damping_time)

    # Construct the array on which the ringdown signal is computed
    # Note: since we are attaching at the *actual* attachment point
    # which will fall *between* grid points we need to add an offset
    # to the ringdown time-series so that ansatze in the ringdown is
    # correctly evaluated.

    t_ringdown = jnp.arange(0, ringdown_time, dt) + (t_match + dt - t_attach)
    t_ringdown_55 = jnp.arange(0, ringdown_time, dt) + (
        t_match_55 + dt - (t_attach - 10)
    )

    # Get the fits for the MR ansatze
    MR_fits = MergerRingdownFits(m1, m2, [0.0, 0.0, chi1], [0.0, 0.0, chi2])
    fits_dict = dict(
        c1f=MR_fits.c1f(), c2f=MR_fits.c2f(), d1f=MR_fits.d1f(), d2f=MR_fits.d2f()
    )

    # Placeholder for the IMR modes. Note that by construction
    # this is longer than is needed for the (5,5) mode, since idx_55<idx
    h = np.zeros(idx + 1 + int(ringdown_time // dt) + 10, dtype=np.complex128)
    N_interp = 5
    for ell_m, mode in hlms_for_compute.items():
        if ell_m == (5, 5):
            t_a = t_attach - 10
            idx_end = idx_55
            t_ring = t_ringdown_55
        else:
            t_a = 1 * t_attach
            idx_end = idx
            t_ring = t_ringdown
        ell, m = ell_m

        amp = jnp.abs(mode)
        phase = jnp.unwrap(jnp.angle(mode))

        idx_interp = np.argmin(np.abs(t_for_compute - t_a))
        left = np.max((0, idx_interp - N_interp))
        right = np.min((idx_interp + N_interp, len(t_for_compute)))

        intrp_amp = InterpolatedUnivariateSpline(
            t_for_compute[left:right], amp[left:right]
        )
        intrp_phase = InterpolatedUnivariateSpline(
            t_for_compute[left:right], phase[left:right]
        )
        amp_max = intrp_amp(t_a)
        damp_max = intrp_amp.derivative()(t_a)
        phi_match = intrp_phase(t_a)
        omega_max = intrp_phase.derivative()(t_a)

        # To improve the stability of the merger-ringdown for odd-m configurations
        # with a minimum in the amplitude close to the attachment point,
        # we directly use the Input Value fits for the frequency,
        # instead of reading its value from the inspiral phase.
        # If the NQCs were *not* applied this would lead to a
        # discontinuity, and one would need to go back to the
        # previous prescription.

        if m % 2 == 1:
            IVfits = InputValueFits(m1, m2, [0.0, 0.0, chi1], [0.0, 0.0, chi2])
            omega_max = IVfits.omega()[ell, m]

        attach_params = dict(
            amp=amp_max,
            damp=damp_max,
            omega=omega_max,
            final_mass=final_mass,
            final_spin=final_spin,
        )

        hring = compute_MR_mode_free(
            t_ring,
            m1,
            m2,
            chi1,
            chi2,
            attach_params,
            ell,
            m,
            fits_dict,
            f_nyquist,
            lmax_nyquist,
            t_match=0,
            phi_match=phi_match,
            qnm_rotation=qnm_rotation,
        )

        # Construct the full IMR waveform
        hIMR[(ell, m)] = 1 * h
        hIMR[(ell, m)][: idx_end + 1] = hlms[(ell, m)][: idx_end + 1]
        hIMR[(ell, m)][idx_end + 1 : idx_end + 1 + len(hring)] = hring[:]

    idx_end = idx

    # Now handle mixed modes
    for ell_m in mixed_modes:
        ell, m = ell_m
        hring = compute_mixed_mode(
            m1,
            m2,
            chi1,
            chi2,
            ell,
            m,
            t_for_compute,
            hlms_for_compute,
            final_mass,
            final_spin,
            t_attach,
            t_ringdown,
            fits_dict,
            f_nyquist,
            lmax_nyquist,
            qnm_rotation=qnm_rotation,
        )
        # Construct the full IMR waveform
        hIMR[(ell, m)] = 1 * h
        hIMR[(ell, m)][: idx_end + 1] = hlms[(ell, m)][: idx_end + 1]
        hIMR[(ell, m)][idx_end + 1 : idx_end + 1 + len(hring)] = hring[:]

    t_IMR = np.arange(len(hIMR[(2, 2)])) * dt
    peak = np.argmax(np.abs(hIMR[(2, 2)]))
    t_IMR -= t_IMR[peak]
    return t_IMR, hIMR


def compute_mixed_mode(
    m1,
    m2,
    chi1,
    chi2,
    ell,
    m,
    t,
    modes,
    final_mass,
    final_spin,
    t_match,
    t_ringdown,
    fits_dict,
    f_nyquist,
    lmax_nyquist,
    qnm_rotation=0.0,
):
    """
    Computes the (3,2) and (4,3) modes, including mode-mixing in the ringdown.

    See Sec. II C of the [SEOBNRv5HM-notes]_ , especially Eqs.(71, 72)

    Args:
        m1 (float): mass of the primary
        m2 (float): mass of the secondary
        chi1 (float): dimensionless spin of the primary
        chi2 (float): dimensionless spin of the secondary
        ell (int): ell index of the desired mode
        m (int): m index of the desired mode
        t (np.ndarray): inspiral time array
        modes (dict): dictionary containing the waveform modes
        final_mass (float): mass of the remnant
        final_spin (float): dimensionless spin of the remnant
        t_match (float): inspiral time at which the merger-ringdown waveform is attached
        t_ringdown (np.ndarray): ringdown time array
        fits_dict (dict): dictionary of fit coefficients in the ringdown anzatz
        f_nyquist (float): Nyquist frequency, needed for checking that RD frequency is resolved
        lmax_nyquist (int): Determines for which modes the nyquist test is applied for
        qnm_rotation (float): Factor rotating the QNM mode frequency in the co-precessing frame (Eq. 33 of Hamilton et al.)


    Returns:
        np.ndarray: the merger-ringdown waveform for the mixed modes

    """
    # Get spheroidal input values
    # These are constructed from spherical input values

    # Spherical modes in the *inspiral*
    mode_lm = modes[ell, m]
    mode_mm = modes[m, m]

    # If the inspiral spherical (ell,m) mode vanishes,
    # we will also set to 0 the merger-ringdown
    # of the (ell,m) spherical mode and return immediately
    if np.max(np.abs(mode_lm)) < 1e-8:
        hring = np.zeros_like(t_ringdown)
        return hring

    # We must ensure continuity, without necessarily assuming NQCs
    # Thus we compute the values at the matching point via
    # interpolation of the inspiral modes.
    # If the NQCs *are* applied then these values would be identical
    # to NQC input values

    # First the (ell,m) mode
    idx_match = np.argmin(np.abs(t - t_match))
    N = 5
    left = np.max((0, idx_match - N))
    right = np.min((idx_match + N, len(t)))
    amp = jnp.abs(mode_lm)
    phase = jnp.unwrap(jnp.angle(mode_lm))
    intrp_amp = InterpolatedUnivariateSpline(t[left:right], amp[left:right])
    intrp_phase = InterpolatedUnivariateSpline(t[left:right], phase[left:right])
    h = intrp_amp(t_match)
    hd = intrp_amp.derivative()(t_match)
    phi_lm = intrp_phase(t_match)
    om = intrp_phase.derivative()(t_match)

    # Now the (m,m) mode
    amp = jnp.abs(mode_mm)
    phase = jnp.unwrap(jnp.angle(mode_mm))
    intrp_amp = InterpolatedUnivariateSpline(t[left:right], amp[left:right])
    intrp_phase = InterpolatedUnivariateSpline(t[left:right], phase[left:right])
    h_mm = intrp_amp(t_match)
    hd_mm = intrp_amp.derivative()(t_match)
    phi_mm = intrp_phase(t_match)
    om_mm = intrp_phase.derivative()(t_match)

    # To improve the stability of the merger-ringdown for configurations
    # with a minimum in the amplitude close to the attachment point,
    # we directly use the Input Value fits for the frequency,
    # instead of reading its value from the inspiral phase.
    # If the NQCs were *not* applied this would lead to a
    # discontinuity, and one would need to go back to the
    # previous prescription.

    if m % 2 == 1:
        IVfits = InputValueFits(m1, m2, [0.0, 0.0, chi1], [0.0, 0.0, chi2])
        om = IVfits.omega()[ell, m]
        om_mm = IVfits.omega()[m, m]

    # Spherical mode we need in the ringdown
    attach_params = dict(
        amp=h_mm,
        damp=hd_mm,
        omega=om_mm,
        final_mass=final_mass,
        final_spin=final_spin,
    )

    hmm_spherical_ringdown = compute_MR_mode_free(
        t_ringdown,
        m1,
        m2,
        chi1,
        chi2,
        attach_params,
        m,
        m,
        fits_dict,
        f_nyquist,
        lmax_nyquist,
        t_match=0 * t_match,
        phi_match=phi_mm,
        qnm_rotation=qnm_rotation,
    )

    # Approximation to spheroidal
    mixing_coeff_mm = np.conj(mu(m, m, m, final_spin))
    hmm_spheroidal = hmm_spherical_ringdown / mixing_coeff_mm

    # Now compute the spheroidal inputs we need

    # Ampltidue at peak
    h_ellm0 = h_ellm0_nu(ell, m, final_spin, h, h_mm, phi_lm, phi_mm)
    # Phase at peak
    ph_ellm0 = phi_ellm0(ell, m, final_spin, h, h_mm, phi_lm, phi_mm)

    # Time derivative of amplitude at peak
    hd_ellm0 = hdot_ellm0_nu(
        ell,
        m,
        final_spin,
        h,
        h_mm,
        hd,
        hd_mm,
        om,
        om_mm,
        phi_lm,
        phi_mm,
    )
    # Frequency at peak
    om_ellm0 = omega_ellm0(
        ell,
        m,
        final_spin,
        h,
        h_mm,
        hd,
        hd_mm,
        om,
        om_mm,
        phi_lm,
        phi_mm,
    )

    attach_params.update(
        amp=h_ellm0,
        damp=hd_ellm0,
        omega=om_ellm0,
        final_mass=final_mass,
        final_spin=final_spin,
    )
    # Compute the coefficients+ansatze for spheroidal mode
    hlm_spheroidal_ringdown = compute_MR_mode_free(
        t_ringdown,
        m1,
        m2,
        chi1,
        chi2,
        attach_params,
        ell,
        m,
        fits_dict,
        f_nyquist,
        lmax_nyquist,
        t_match=0 * t_match,
        phi_match=ph_ellm0,
        qnm_rotation=qnm_rotation,
    )
    # Reconstruct the spherical mode
    hring = hmm_spheroidal * np.conj(
        mu(m, ell, m, final_spin)
    ) + hlm_spheroidal_ringdown * np.conj(mu(m, ell, ell, final_spin))

    return hring


def NQC_correction(
    inspiral_modes: Dict,
    t_modes: np.ndarray,
    polar_dynamics: np.ndarray,
    t_peak: float,
    nrDeltaT: float,
    m_1: float,
    m_2: float,
    chi_1: float,
    chi_2: float,
):
    """Given the inspiral modes and the dynamics this function
    computes the NQC coefficients at t_peak-nrDeltaT

    Args:
        inspiral_modes (Dict): Dictionary of inspiral modes (interpolated)
        t_modes (np.ndarray): Time array for inspiral modes
        polar_dynamics (np.ndarray): Dynamics array from ODE solver (unequally spaced)
        t_peak (float): The time of the peak of the orbital frequency
        nrDeltaT (float): The shift from peak of the orbital frequency to peak of (2,2) mode
        m_1 (float): Mass of the primary
        m_2 (float): Mass of the secondary
        chi_1 (float): z-component of dimensionless spin of the primary
        chi_2 (float): z-component Dimensionless spin of the secondary
    """

    # Compute omega

    r = polar_dynamics[0]
    pr = polar_dynamics[1]
    omega_orb = polar_dynamics[2]
    input_value_fits = InputValueFits(m_1, m_2, [0.0, 0.0, chi_1], [0.0, 0.0, chi_2])
    fits_dict = dict(
        amp=input_value_fits.habs(),
        damp=input_value_fits.hdot(),
        ddamp=input_value_fits.hdotdot(),
        omega=input_value_fits.omega(),
        domega=input_value_fits.omegadot(),
    )

    # Loop over every mode
    nqc_coeffs = {}
    for ell_m, mode in inspiral_modes.items():
        amp = np.abs(mode)
        phase = np.unwrap(np.angle(mode))
        ell, m = ell_m
        if ell == 5 and m == 5:
            # (5,5) mode is special
            extra = -10
        else:
            extra = 0
        # NQC_coeffs = EOBCalculateNQCCoefficientsV4(amp,phase,r,pr,omega_orb,ell,m,t_peak,t_modes,m1,m2,chi1,chi2)
        # For equal mass, non-spinning cases odd m modes vanish, so don't try to compute NQCs
        if (
            m % 2
            and np.abs(m_1 - m_2) < 1e-4
            and np.abs(chi_1) < 1e-4
            and np.abs(chi_2) < 1e-4
        ) or (m % 2 and np.abs(m_1 - m_2) < 1e-4 and np.abs(chi_1 - chi_2) < 1e-4):
            continue

        else:

            # Compute the NQC coeffs
            NQC_coeffs = EOBCalculateNQCCoefficients_freeattach(
                amp,
                phase,
                r,
                pr,
                omega_orb,
                ell,
                m,
                t_peak,
                t_modes,
                m_1,
                m_2,
                chi_1,
                chi_2,
                nrDeltaT - extra,
                fits_dict,
            )

            NQC_coeffs["a3S"] = 0
            NQC_coeffs["a4"] = 0
            NQC_coeffs["a5"] = 0
            NQC_coeffs["b3"] = 0
            NQC_coeffs["b4"] = 0

        nqc_coeffs[(ell, m)] = deepcopy(NQC_coeffs)

    return nqc_coeffs


def apply_nqc_corrections(
    hlms: Dict[Any, Any], nqc_coeffs: Dict[Any, Any], polar_dynamics: np.ndarray
):
    """
    Loop over modes and multiply them by NQC correction

    Args:
        hlms (Dict): Dictionary of inspiral modes
        nqc_coeffs (Dict): Dictionary of NQC coefficients
        polar_dynamics (np.ndarray): Dynamics array

    """
    r, pr, omega_orb = polar_dynamics
    for key in hlms.keys():
        ell, m = key
        try:
            NQC_coeffs = nqc_coeffs[(ell, m)]
        except KeyError:
            continue
        correction = EOBNonQCCorrection(r, None, pr, None, omega_orb, NQC_coeffs)
        hlms[key] *= correction



#!/usr/bin/env python
"""
Contains functions associated with evolving the equations of motion
"""

import numpy as np
from scipy.interpolate import CubicSpline
from scipy.signal import argrelmin
from numba import jit

import pygsl_lite.errno as errno
import pygsl_lite.odeiv2 as odeiv2

from .initial_conditions_aligned_opt import computeIC_opt
from .rhs_aligned import get_rhs, augment_dynamics


step = odeiv2.pygsl_lite_odeiv2_step
_control = odeiv2.pygsl_lite_odeiv2_control
evolve = odeiv2.pygsl_lite_odeiv2_evolve


class control_y_new(_control):
    def __init__(self, eps_abs, eps_rel):
        a_y = 1
        a_dydt = 1
        super().__init__(eps_abs, eps_rel, a_y, a_dydt, None)


@jit(nopython=True)
def h_max(r):
    return 1


def ODE_system_RHS_opt(t: float, z: np.ndarray, args) -> np.ndarray:
    """Return the dynamics equations for aligned-spin systems

    Args:
        t (float): The current time
        z (np.array): The dynamics variables, stored as (q,p)
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use. Must have same signature as Hamiltonian
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary

    Returns:
        np.array: The dynamics equations, including RR
    """
    return get_rhs(t, z, *args)


def compute_dynamics_opt(
    omega0,
    H,
    RR,
    chi_1,
    chi_2,
    m_1,
    m_2,
    rtol=1e-11,
    atol=1e-12,
    backend="solve_ivp",
    params=None,
    step_back=100,
    max_step=0.1,
    min_step=1.0e-9,
    y_init=None,
    r_stop=None,
):
    """
    Main function to integrate the dynamics
    The RHS of the equations are given in Eq(2) of [Buades2021]_ .
    See rhs_aligned.pyx for more details.
    Uses GSL Dormand-Prince 8th order integrator.
    Args:
        omega0 (float): initial orbital frequency
        H (Hamiltonian): The Hamiltonian object to use
        RR (function): The RR force to use. Must have same signature as Hamiltonian
        chi_1 (float): z-component of the primary spin
        chi_2 (float): z-component of the secondary spin
        m_1 (float): mass of the primary
        m_2 (float): mass of the secondary
        rtol (float): relative tolerance
        atol (float): absolute tolerance
        step_back (float): step back for the start of the fine dynamics
        r_stop (float): minimum final separation for the dynamics

    Returns:
        np.array, np.array: coarse and fine dynamics arrays
    """

    sys = odeiv2.system(
        ODE_system_RHS_opt, None, 4, [H, RR, chi_1, chi_2, m_1, m_2, params]
    )

    T = odeiv2.step_rk8pd
    s = step(T, 4)
    c = control_y_new(atol, rtol)
    e = evolve(4)

    t = 0
    t1 = 2.0e9
    if r_stop < 0:
        r_stop = 1.4

    if y_init is None:
        r0, pphi0, pr0 = computeIC_opt(
            omega0, H, RR, chi_1, chi_2, m_1, m_2, params=params
        )
        y0 = np.array([r0, 0.0, pr0, pphi0])
    else:
        y0 = y_init.copy()
    y = y0
    if y_init is None:
        h = 2 * np.pi / omega0 / 5
    else:
        h = 0.5
    omega_previous = omega0
    res_gsl = []
    ts = []
    omegas = []
    ts.append(0.0)
    res_gsl.append(y)

    p_circ = np.zeros(2)
    peak_omega = False
    peak_pr = False
    while t < t1:
        # Take a step
        status, t, h, y = e.apply(c, s, sys, t, t1, h, y)
        if status != errno.GSL_SUCCESS:
            print("break status", status)
            break
        # Compute the error for the step controller
        e.get_yerr()

        # Append the last step
        res_gsl.append(y)
        ts.append(t)

        # Compute the RHS after the step is done

        r = y[0]

        # Check if the proposed step is larger than the maximum timestep
        # h_mx = h_max(r)
        # h = h / h_mx

        # Handle termination conditions
        if r <= 6:
            deriv = ODE_system_RHS_opt(t, y, [H, RR, chi_1, chi_2, m_1, m_2, params])
            drdt = deriv[0]
            omega = deriv[1]
            dprdt = deriv[2]
            """
            h_small = np.max((0.01,2*np.pi/(2.*omega) / (1 + np.exp(-(r - 4) / 0.13))))
            if h > h_small:
                h = h_small
            """

            if omega < omega_previous:
                peak_omega = True
                break
            if drdt > 0:
                break
            if dprdt > 0:
                peak_pr = True
                break
            if r <= r_stop:
                break
            if r < 3:
                q_vec = y[:2]
                p_circ[1] = y[-1]
                omega_circ = H.omega(q_vec, p_circ, chi_1, chi_2, m_1, m_2)
                if omega_circ > 1:
                    break
            omega_previous = omega

    ts = np.array(ts)
    dyn = np.array(res_gsl)

    if peak_omega:
        t_desired = ts[-1] - step_back - 50
    else:
        t_desired = ts[-1] - step_back

    idx_close = np.argmin(np.abs(ts - t_desired))
    if ts[idx_close] > t_desired:
        idx_close -= 1

    # Guard against the case where when using PA dynamics,
    # there is less than step_back time between the start
    # of the ODE integration and the end of the dynamics
    # In that case make the fine dynamics be _all_ dynamics
    # except the 1st element
    if t_desired < ts[1]:
        idx_close = 1
        step_back = ts[-1] - ts[idx_close]
    dyn_coarse = np.c_[ts[:idx_close], dyn[:idx_close]]
    dyn_fine = np.c_[ts[idx_close:], dyn[idx_close:]]
    """
    print(f"t_desired={t_desired}")
    print(f"len(dyn_coarse)={len(dyn_coarse)}")
    print(f"len(dyn_fine)={len(dyn_fine)}")
    print(f"End time: {ts[-1]}")
    print(f"idx_close:{idx_close}, t[idx_close]={ts[idx_close]}")
    """
    dyn_coarse = augment_dynamics(dyn_coarse, chi_1, chi_2, m_1, m_2, H)
    dyn_fine = augment_dynamics(dyn_fine, chi_1, chi_2, m_1, m_2, H)
    t_peak = None
    if peak_omega:
        intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, -2])
        left = dyn_fine[0, 0]
        right = dyn_fine[-1, 0]
        t_peak = iterative_refinement(intrp.derivative(), [left, right])

    if peak_pr:
        intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, 3])
        left = dyn_fine[-1, 0] - 10
        right = dyn_fine[-1, 0]
        t_peak = iterative_refinement(intrp.derivative(), [left, right], pr=True)

    dyn_fine = interpolate_dynamics(
        dyn_fine[:, :-3], peak_omega=t_peak, step_back=step_back
    )
    dyn_fine = augment_dynamics(dyn_fine, chi_1, chi_2, m_1, m_2, H)

    return dyn_coarse, dyn_fine


def iterative_refinement(f, interval, levels=2, dt_initial=0.1, pr=False):
    """
    Attempts to find the peak of Omega/pr iteratively.
    Needed to ensure accurate attachment when the attachment point is the last point of the dynamics.

    Args:
        f (PPoly): derivative of the splined dynamics
        interval (list): interval where to look for the peak
        levels (int): number of iterations
        dt_initial (float): initial time step
        pr (bool): whether the end of the dynamics is due to a peak of pr instead of Omega

    Returns:
        np.array: interpolated dynamics array

    """
    left = interval[0]
    right = interval[1]
    for n in range(1, levels + 1):
        dt = dt_initial / (10**n)
        t_fine = np.arange(interval[0], interval[1], dt)
        deriv = np.abs(f(t_fine))

        mins = argrelmin(deriv, order=3)[0]
        if len(mins) > 0:
            result = t_fine[mins[0]]

            interval = max(result - 10 * dt, left), min(result + 10 * dt, right)

        else:
            if pr:
                return interval[-1]
            else:
                return (interval[0] + interval[-1]) / 2
    return result


def interpolate_dynamics(dyn_fine, dt=0.1, peak_omega=None, step_back=250.0):
    """
    Interpolate the dynamics to a finer grid.
    This replaces stepping back that was used in older EOB models.

    Args:
        dyn_fine (np.array): dynamics array
        dt (float): time step to which to interpolate
        peak_omega (float): position of the peak (stopping condition for the dynamics)
        step_back (float): step back relative to the end of the dynamics

    Returns:
        np.array: interpolated dynamics array

    """

    res = []
    n = len(dyn_fine)

    if peak_omega:
        start = max(peak_omega - step_back, dyn_fine[0, 0])
        t_new = np.arange(start, peak_omega, dt)

    else:
        t_new = np.arange(dyn_fine[0, 0], dyn_fine[-1, 0], dt)

    for i in range(1, dyn_fine.shape[1]):
        intrp = CubicSpline(dyn_fine[:, 0], dyn_fine[:, i])
        res.append(intrp(t_new))

    res = np.array(res)
    res = res.T
    return np.c_[t_new, res]



"""
Fits for the calibration coefficients in the Hamiltonian.
"""

# This file was automatically generated by polynomial_fit_v4.py on 2022-09-01 18:37:25.690079
# running at revision 159a400f4d78ddb6f88aa7bd43c2d31289a3a4ae.
# The posterior data used was spinning_pst_handcut.h5 with md5sum: 3cc8d7aab62dbfbb451f57548b6d1522
#
# The RMSE was [37.02442072  2.49453561]
#
# The TPL residuals were [-16.73657386 -12.40339386   1.35879203  -3.05734409]
#
# The settings used:
#
# input_physical_parameters = [ "q", "chi1", "chi2",]
# reduction = "median"
# input_transformations = "ETA_CHIEFF_CHIA"
# calibration_parameters = [ "dSO", "NR_deltaT",]
# calibration_labels = [ "$\\Delta t_{22}$", "$d_\\mathrm{SO}$",]
# force_independent = false
# TPL_rescaling = true
# k = 0
# fitk = false
# hierarchical = true
# hierarchical_function = "dt22_non_spinning_fit"
# weight_function = "seobnrv4"
# regularize = false
# alpha = 0.8
# lambda = 0.3
# solver = "SLSQP"
# n_restarts = 512
#
# [feature_map]
# x0 = "nu"
# x1 = "ap"
# x2 = "am"
#
# [feature_array]
# dSO = [ "1", "x0", "x0*x0", "x0*x0*x0", "x0*x0*x1", "x0*x0*x2", "x0*x1", "x0*x1*x1", "x0*x1*x2", "x0*x2", "x0*x2*x2", "x1", "x1*x1", "x1*x1*x1", "x1*x1*x2", "x1*x2", "x1*x2*x2", "x2", "x2*x2", "x2*x2*x2",]
# NR_deltaT = [ "x0*x0*x1", "x0*x0*x2", "x0*x1", "x0*x1*x1", "x0*x1*x2", "x0*x2", "x0*x2*x2", "x1", "x1*x1", "x1*x1*x1", "x1*x1*x2", "x1*x2", "x1*x2*x2", "x2", "x2*x2", "x1*x1*x1*x1",]
#
# [extra_physical_info]
# TPL = false
# additional_terms = "spinning_TPL_info"
# additional_terms_weight = 25.0


import numpy as np
from numba import jit


@jit(nopython=True)
def dSO(nu, ap, am):
    """
    Fit for the spin-orbit calibration coefficient dSO at 4.5 PN

    See Eq(91) in [SEOBNRv5HM-notes]_ .

    Args:
        nu (float): reduced mass ratio.
        ap (float): symmetric spin combination ap = m1*chi1 + m2*chi2.
        am (float): antisymmetric spin combination ap = m1*chi1 - m2*chi2.

    Returns:
        float: The value of dSO.

    """
    return (
        -7.71251231383957 * am ** 3
        - 17.2294679794015 * am ** 2 * ap
        - 238.430383378296 * am ** 2 * nu
        + 69.5461667822545 * am ** 2
        - 10.5225438990315 * am * ap ** 2
        + 362.767393298729 * am * ap * nu
        - 85.8036338010274 * am * ap
        - 1254.66845939312 * am * nu ** 2
        + 472.431937787377 * am * nu
        - 39.742317057316 * am
        - 7.58458103577458 * ap ** 3
        - 42.7601129678844 * ap ** 2 * nu
        + 18.1783435552183 * ap ** 2
        - 201.905934468847 * ap * nu ** 2
        - 90.5790079104259 * ap * nu
        + 49.6299175121658 * ap
        + 478.546231305475 * nu ** 3
        + 679.521769948995 * nu ** 2
        - 177.334831768076 * nu
        - 37.6897780220529
    )


@jit(nopython=True)
def NR_deltaT(nu, ap, am):
    """
    Fit for the time difference from reference point to the attachment time, spinning part

    See Eq(90) in [SEOBNRv5HM-notes]_ .
    The NR_deltaT is done hierarchically, and one needs to add the non-spinning and spinning contributions:
    NR_deltaT = NR_deltaT_NS(nu) +  NR_deltaT(nu, ap, am)

    Args:
        nu (float): reduced mass ratio.
        ap (float): symmetric spin combination ap = m1*chi1 + m2*chi2.
        am (float): antisymmetric spin combination ap = m1*chi1 - m2*chi2.

    Returns:
        float: The value of NR_deltaT.
    """
    return nu ** (-1.0 / 5 + 0 * nu) * (
        8.39238879807543 * am ** 2 * ap
        - 16.9056858928167 * am ** 2 * nu
        + 7.23410583477034 * am ** 2
        + 6.38975598319936 * am * ap ** 2
        + 179.569824846781 * am * ap * nu
        - 40.6063653476775 * am * ap
        + 144.253395844761 * am * nu ** 2
        - 90.1929138487509 * am * nu
        + 14.2203101910927 * am
        - 6.78913884987037 * ap ** 4
        + 5.39962303470497 * ap ** 3
        - 132.224950777226 * ap ** 2 * nu
        + 49.8016443361381 * ap ** 2
        + 384.201018794943 * ap * nu ** 2
        - 141.253181790353 * ap * nu
        + 17.5710132409988 * ap
    )


"""
Non-spinning functions obtained by a least-square fit of the maxL of the calibration posteriors
"""


@jit(nopython=True)
def a6_NS(nu):
    """
    Fit for the non-spinning calibration coefficient a6 at 5PN in A potential.

    See Eq(88) in [SEOBNRv5HM-notes]_ .

    Args:
        nu (float): reduced mass ratio.

    Returns:
        float: The value of a6.
    """
    par = np.array(
        [4.17877875e01, -3.02193382e03, 3.34144394e04, -1.69019140e05, 3.29523262e05]
    )
    return par[0] + par[1] * nu + par[2] * nu ** 2 + par[3] * nu ** 3 + par[4] * nu ** 4


@jit(nopython=True)
def NR_deltaT_NS(nu):
    """
    Fit for the time difference from reference point to the attachment time, no spin part.

    See Eq(89) in [SEOBNRv5HM-notes]_ .
    The NR_deltaT is done hierarchically, and one needs to add the non-spinning and spinning contributions:
    NR_deltaT = NR_deltaT_NS(nu) +  NR_deltaT(nu, ap, am)

    Args:
        nu (float): reduced mass ratio.

    Returns:
        float: The non-spinning value of NR_deltaT.
    """
    par = np.array(
        [1.00513217e01, -5.96231800e01, -1.05687385e03, -9.79317619e03, 5.55652392e04]
    )
    return nu ** (-1.0 / 5 + par[0] * nu) * (
        par[1] + par[2] * nu + par[3] * nu ** 2 + par[4] * nu ** 3
    )
